/*
 * PlaybackController.cpp
 *
 *  Created on: 09.07.2010
 *      Author: darkstar
 */

#include "playbackcontroller.h"

/* Helper */

void swap(int &n1, int &n2)
{
	int n = n1;
	n1 = n2;
	n2 = n;
}

/* PlaybackController */

PlaybackController::PlaybackController()
	:	aspectSource_(NULL),
		playOrder_(ORDER_NORMAL),
		randomPlayOrderTable(NULL),
		count_(0),
		randomPlayOrderTableIndex_(0),
		activeItemIndex_(-1),
		activeItem_(0, 0, NULL)
{
	// TODO Auto-generated constructor stub

}

PlaybackController::~PlaybackController()
{
	// TODO Auto-generated destructor stub
}

void PlaybackController::attachTo(PlayListAspectListViewRenderer *source)
{
	detach();

	if (source)
	{
		aspectSource_ = source;
		count_ = aspectSource_->itemList().count();

		connect(aspectSource_, SIGNAL(updateView()), this, SLOT(updateView()));

		activeItem_.setMediaDatabase(aspectSource_->playListAspect()->playList()->mediaDatabase());

		updateView();
	}
}

void PlaybackController::detach()
{
	ensurePlayListOrderIsValid();

	if (aspectSource_)
	{
		disconnect(aspectSource_, SIGNAL(updateView()), this, SLOT(updateView()));

		activeItemIndex_ = -1;
		count_ = 0;

		if (randomPlayOrderTable)
			delete randomPlayOrderTable;

		randomPlayOrderTable = NULL;

		activeItem_.setMediaDatabase(NULL);
		activeItem_.setID(0);

		aspectSource_ = NULL;
	}
}

void PlaybackController::updateView()
{
	// try to find the activeItemIndex based on our current cached activeItem...
	setActiveItem(activeItem_);
}

void PlaybackController::setPlayOrder(PlayOrder newOrder)
{
	randomPlayOrderTableIndex_ = 0;
	playOrder_ = newOrder;
	// generateRandomPlayOrderTable();
}

void PlaybackController::updateRandomPlayOrderTable()
{
	if (!aspectSource_) return;

	if (!randomPlayOrderTable || count_ != aspectSource_->itemList().count())
		generateRandomPlayOrderTable();
}

void PlaybackController::generateRandomPlayOrderTable()
{
	if (!aspectSource_) return;

	randomPlayOrderTableIndex_ = 0;

	if (randomPlayOrderTable)
		delete randomPlayOrderTable;

	randomPlayOrderTable = NULL;

	if (count_ < 1)
		return;

	if (playOrder_ == ORDER_RANDOM || playOrder_ == ORDER_REPEAT_RANDOM)
	{
		randomPlayOrderTable = new int[count_];

		for (int i = 0; i != count_; ++i)
			randomPlayOrderTable[i] = i;

		// Randomize
		std::srand(time(NULL));

		for (int i = 0; i != count_; ++i)
			swap(randomPlayOrderTable[i], randomPlayOrderTable[std::rand() % count]);

		if (activeItem_)
		{
			int selectedIndex = aspectSource_->itemList().findRef(activeItem_);

			// make selected index the first in our play order table...
			for (int i = 0; i != count_; ++i)
				if (randomPlayOrderTable[i] == selectedIndex)
				{
					swap(randomPlayOrderTable[i], randomPlayOrderTable[0]);
					break;
				}
		}
	}
}

void PlaybackController::setActiveItemIndex(int itemIndex)
{
	activeItemIndex_ = itemIndex;

	if (activeItemIndex_ > -1)
	{
		// copy from the current item in the list...
		activeItem_ = *aspectSource_->itemList().at(activeItemIndex_);
	}

	emit activeItemChanged();
}

int PlaybackController::nextActiveItemIndex(bool modifyRandomPlayOrderTableIndex)
{
	if (!aspectSource_ || count_ < 1)
		return -1;

	int itemIndex = -1;

	switch (playOrder_)
	{
	case ORDER_NORMAL:
		if (activeItemIndex_ == -1)
			itemIndex = getSelectedItem();
		else if (activeItemIndex_ == count_ - 1)
			itemIndex = -1;
		else
			itemIndex = activeItemIndex_ + 1;

		break;

	case ORDER_REPEAT_ONE:
		itemIndex = activeItemIndex_;

		break;

	case ORDER_REPEAT_ALL:
		if (activeItemIndex_ == -1)
			itemIndex = getSelectedItem();
		else if (activeItemIndex_ == count_ - 1)
			itemIndex = 0;
		else
			itemIndex = activeItemIndex_ + 1;

		break;

	case ORDER_RANDOM:
		updateRandomPlayOrderTable();

		if (randomPlayOrderTableIndex_ + 1 < count_)
			itemIndex = randomPlayOrderTable[modifyRandomPlayOrderTableIndex ? ++randomPlayOrderTableIndex_ : randomPlayOrderTableIndex_ + 1];
		else
			itemIndex = -1;

		break;

	case ORDER_REPEAT_RANDOM:
		updateRandomPlayOrderTable();

		if (randomPlayOrderTableIndex_ + 1 < count)
			itemIndex = randomPlayOrderTable[modifyRandomPlayOrderTableIndex ? ++randomPlayOrderTableIndex_ : randomPlayOrderTableIndex_ + 1];
		else
			itemIndex = randomPlayOrderTable[modifyRandomPlayOrderTableIndex ? randomPlayOrderTableIndex_ = 0 : 0];

		break;
	}

	return itemIndex;
}

int PlaybackController::previousActiveItemIndex(bool modifyRandomPlayOrderTableIndex)
{
	if (!aspectSource_ || count_ < 1)
		return -1;

	int itemIndex = -1;

	switch (playOrder_)
	{
	case ORDER_NORMAL:
		if (activeItemIndex_ == -1)
			itemIndex = getSelectedItem();
		else if (activeItemIndex_ == 0)
			itemIndex = -1;
		else
			itemIndex = activeItemIndex_ - 1;

		break;

	case ORDER_REPEAT_ONE:
		itemIndex = activeItemIndex_;

		break;

	case ORDER_REPEAT_ALL:
		if (activeItemIndex_ == -1)
			itemIndex = getSelectedItem();
		else if (activeItemIndex_ == 0)
			itemIndex = count_ - 1;
		else
			itemIndex = activeItemIndex_ -1;

		break;

	case ORDER_RANDOM:
		updateRandomPlayOrderTable();

		if (randomPlayOrderTableIndex_ <= 0)
			itemIndex = -1;
		else
			itemIndex = randomPlayOrderTable[modifyRandomPlayOrderTableIndex ? --randomPlayOrderTableIndex_ : randomPlayOrderTableIndex_ - 1];

		break;

	case ORDER_REPEAT_RANDOM:
		updateRandomPlayOrderTable();

		if (randomPlayOrderTableIndex_ <= 0)
			itemIndex = -1;
		else
			itemIndex = randomPlayOrderTable[modifyRandomPlayOrderTableIndex ? --randomPlayOrderTableIndex_ : randomPlayOrderTableIndex_ - 1];

		break;
	}

	return itemIndex;
}

int PlaybackController::activeItemIndex()
{
	if (!aspectSource_ || count_ < 1)
		return -1;

	int itemIndex = activeItemIndex_;

	if (itemIndex == -1)
	{
		itemIndex = nextActiveItemIndex(true);
		setActiveItemIndex(itemIndex);
	}

	return itemIndex;
}

void PlaybackController::setActiveItem(const PlayListItem &item)
{
	if (!aspectSource_)
		return;

	// try to find activeItem in the itemList and set activeItemIndex
	int mediaIDmatchIndex = -1;
	int newActiveItemIndex = -1;

	const PlayListItemList &itemList = aspectSource_->itemList();

	for (int i = 0; i < count_; ++i)
	{
		PlayListItem *listItem = itemList.at(i);

		if (listItem->mediaID() == item.mediaID())
			mediaIDmatchIndex = i;

		if (listItem->id() == item.id() && mediaIDmatchIndex == i)
		{
			newActiveItemIndex = i;
			break;
		}
	}

	if (newActiveItemIndex == -1 && mediaIDmatchIndex > -1)
		newActiveItemIndex = mediaIDmatchIndex;

	setActiveItemIndex(newActiveItemIndex);
}

const PlayListItem &PlaybackController::activeItem()
{
	int index = activeItemIndex();

	if (index > -1)
		return aspectSource_->itemList().at(index);
	else
		return activeItem_;
}

const PlayListItem &PlaybackController::nextItem(bool modifyRandomPlayOrderTableIndex)
{
	int index = nextActiveItemIndex(modifyRandomPlayOrderTableIndex);

	if (index > -1)
		return aspectSource_->itemList().at(index);
	else
		return NULL;
}

const PlayListItem &PlaybackController::previousItem(bool modifyRandomPlayOrderTableIndex)
{
	int index = nextActiveItemIndex(modifyRandomPlayOrderTableIndex);

	if (index > -1)
		return aspectSource_->itemList().at(index);
	else
		return NULL;
}

void PlaybackController::startOrPauseCurrentPlaybackOrStartNewPlayback()
{
	startOrPauseCurrentPlaybackOrStartNewPlayback(false);
}

void PlaybackController::startOrPauseCurrentPlaybackOrStartNewPlayback(bool forcePlayNewFile)
{
	// Toggle play/pause
	if (mplayer_->isPlaying() && !forcePlayNewFile)
	{
		if (mplayer_->isPaused())
		{
			mplayer_->enableTemporaryOSD();

			mplayer_->pause();
			playListView_->setActiveItemPaused(false);
			operationPanel_->setPlayPause(false);

			emit playbackUnpaused();

			osdTimer_->start(3000, true);
		}
		else
		{
			mplayer_->enableTemporaryOSD();

			mplayer_->pause();
			playListView_->setActiveItemPaused(true);
			operationPanel_->setPlayPause(true);

			emit playbackPaused();
		}
	}
	// Start mplayer
	else if (playListView_->childCount() > 0 && playListView_->currentItem() != NULL)
	{
		generateRandomPlayOrderTable();
		internalStartPlayback();
		Media *media = playListView_->currentMedia();
		if (media->isAudio() && qConfig.automaticallySwitchToPlayInfoEnabled) {
			// Show info widget
			setActiveView(Quasar::ViewPlayInfo);
		}

#ifdef QTOPIA
		if (qConfig.isBatteryPlusEnabled) {
			if (media->isVideo()) {
				g_BatteryPlus.setMode(qConfig.batteryPlusMode4Video);
			} else {
				g_BatteryPlus.setMode(qConfig.batteryPlusMode4Audio);
			}
		}

		qPlatformServices.disableSuspend();
#endif
	}
}

void PlaybackController::stopCurrentPlaybackAndPlayCurrentItem()
{
	DENTERMETHOD;

	PlayListViewItem *item = static_cast<PlayListViewItem *>(playListView_->currentItem());

	if (!item)
		return;

	if (item->media()->errorDetected())
	{
		TextViewDialog errDialog(tr("MPlayer Error Message"),
			 tr("MPlayer could not be started for the following reason:"),
			 item->media()->lastError(),
			 tr("Perhaps your format configuration is wrong.") + "\n" +
			 tr("Do you want to open the configuration dialog?"),
			 tr("Yes"),
			 tr("No"),
			 this
		);

		if (errDialog.exec())
			configuration(true);
	}
	else
	{
		internalStopPlayback(true, false);
		playListView_->setActiveItem(item);
		playListView_->setActiveItemPaused(false);
		qApp->processEvents();
		startOrPauseCurrentPlaybackOrStartNewPlayback(true);
	}

	DEXITMETHOD;
}

void PlaybackController::stopCurrentPlayback()
{
	setActiveView(Quasar::ViewLastActivePlaylist);
	internalStopPlayback();

#ifdef QTOPIA
	qPlatformServices.enableSuspend();

	if (qConfig.isBatteryPlusEnabled)
		g_BatteryPlus.setMode(g_BatteryPlus.initialMode());
#endif
}

void PlaybackController::internalStartPlayback()
{
	Media *media = playListView_->currentMedia();

	emit preStartPlayback();

	DPRINTF("Current Media: %s", (const char *)(media->location().toString().toUtf8()));

	nextMediaPrefetched_ = false;
	currentMediaLength_ = media->length();

	// Start mplayer
	if (mplayer_->play(media) == MPlayer::InitializationFailed)
	{
		int result = QMessageBox::critical(this, tr("mplayer"),
				 tr("mplayer could not be executed or\nwas not found on your system.\nWould you like to specify the location\nof the mplayer binary on your system?"),
				 QMessageBox::Yes | QMessageBox::Default, QMessageBox::No | QMessageBox::Escape);

		if (result == QMessageBox::Yes)
		{
			configuration(true);
		}

		return;
	}

	// Set Operation Panel
	operationPanel_->setPlayPause(false);
	operationPanel_->setTotalPlayTime(currentMediaLength_);

	playListView_->setActiveItemPaused(false);

	if (media->isVideo())
	{
#ifndef QTOPIA
		if (qConfig.isOverlayEnabled)
		{
#endif
			videoPlaybackWindow_->hide();
			videoPlaybackWindow_->reparent(0, QPoint(0, 0));
			videoPlaybackWindow_->resize(qApp->desktop()->size());
			videoPlaybackWindow_->setCursor(QCursor(Qt::blankCursor));
			videoPlaybackWindow_->showFullScreen();
#ifndef QTOPIA
		}
#endif
	}
	else if (media->isAudio())
	{
		playInfo_->setMediaInfo(*playListView_->currentMedia());

		// Get current number
		int n = playListView_->itemIndex(playListView_->activeItem()) + 1;
		playInfo_->setTrackNumbers(n, playListView_->childCount());
	}

	emit playbackStarted(media);
}

void PlaybackController::internalStopPlayback(bool temporarily, bool stopBackend)
{
	// Set Panels
	if (!temporarily)
	{
		operationPanel_->setCurrentPlayTime(0);
		operationPanel_->setPlayPause(true);

		playInfo_->clear();
	}

	// Stop mplayer
	emit playbackStopped();

	if (stopBackend)
		mplayer_->stopCurrentPlayback();

	emit playbackStopped();

	if (!temporarily)
	{
		playListView_->setActiveItem(NULL);
		playListView_->setActiveItemPaused(false);
	}

	if (videoPlaybackWindow_->isVisible())
	{
		videoPlaybackWindow_->hide();
		videoPlaybackWindow_->reparent(this, QPoint(0, 0));
		videoPlaybackWindow_->resize(QSize(0, 0));
		videoPlaybackWindow_->unsetCursor();
	}
}

void PlaybackController::playNextItem()
{
	if (int index = nextActiveItemIndex(true) > -1)
	{
		++keyEventCount_;
		qApp->processEvents();
		--keyEventCount_;

		if (keyEventCount_ <= 0)
		{
			internalStopPlayback(true, false);
			setActiveItemIndex(index);
			internalStartPlayback();
			keyEventCount_ = 0;
		}
	}
	else
	{
		stopCurrentPlayback();
	}
}

void PlaybackController::playPreviousItem()
{
	if (int index = previousActiveItemIndex(true) > -1)
	{
		++keyEventCount_;
		qApp->processEvents();
		--keyEventCount_;

		if (keyEventCount_ <= 0)
		{
			internalStopPlayback(true, false);
			setActiveItemIndex(index);
			internalStartPlayback();
			keyEventCount_ = 0;
		}
	}
}
